#!/usr/bin/perl -w

use strict qw(subs vars);
use Cwd;
use File::Basename;
use Date::Calc;
use LWP::UserAgent;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my %http_info;
my %files;

my $www_data_dir     = '/var/www/fci/';
my $xml_data_dir     = 'xml-data/';

my %cfg = ( self                       => 'fci-update-jmol-projects.pl',
            version                    => '0.1.1',
	    
	    verbose                    => 0,
	    quiet                      => 0,
	    download                   => 1,
	    force                      => 0,

	    'www-data-dir'             => $www_data_dir,
	    'xml-data-dir'             => $xml_data_dir,
	    'jmol-projects-file'       => $xml_data_dir.'jmol-projects.xml',
	    'xml-project-summary-file' => $xml_data_dir.'project-summary.xml',

	    'url'                      => 'http://jmol.sourceforge.net/fah/fah-projects.xml',

          );


# Get commandline parameters
Getopt::Mixed::init("dir=s
                     url=s
                     no-download
		     force
                     verbose
		     quiet
		     help
		     d>dir
		     u>url
		     n>no-download
		     f>force
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'url'}          = $value if($option eq 'url');
	
	$cfg{download}   = 0      if($option eq 'no-download');
	$cfg{force}      = 1      if($option eq 'force');
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if force was requested, override no-download
if($cfg{force}) {
	$cfg{download} = 1;
}
# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
if(!-e $cfg{'jmol-projects-file'}) {
	print "Creating jmol-projects XML file.\n" if($cfg{verbose});
	my ($success, $reason) = save_file($cfg{'jmol-projects-file'}, "", mode => '0664');

	if(!$success) {
		print $reason;
		exit 1;
	}
}
if(!-r $cfg{'jmol-projects-file'}) {
	print "Error: Cannot read jmol-projects.xml file!\n";
	exit 1;
}
if(!-r $cfg{'xml-project-summary-file'}) {
	print "Error: Cannot read project-summary.xml file!\n";
	exit 1;
}


# read the xml files
my $jmol_projects   = read_xml(%cfg, Name => 'jmol-projects') || die "Error: Cannot read jmol-projects XML file!\n";
my $project_summary = read_xml(%cfg, Name => 'xml-project-summary') || die "Error: Cannot read project-summary XML file!\n";

#print Dumper($jmol_projects);
#print Dumper($project_summary);

# create useragent object which will handle the HTTP connections
my $ua  = LWP::UserAgent->new();
   $ua->agent("$cfg{self}/$cfg{version}");


# download the XML file
# http://jmol.sourceforge.net/fah/?project=2124

# get the Modified HTTP header value from the first line of the file
print "Looking up Last-Modified HTTP header ... " if($cfg{verbose});

my $last_modified = '';
if($jmol_projects->{fah_projects}->{last_modified} &&
   $jmol_projects->{fah_projects}->{last_modified} !~ /^\s*$/
) {
        $last_modified = $jmol_projects->{fah_projects}->{last_modified};

        print "Found.\nLast-Modified: $last_modified\n" if($cfg{verbose});
}
else {
        print "Not found.\n" if($cfg{verbose});
}
print "\n" if($cfg{verbose});


# check the URL for modification, only download if we have an older version
if($last_modified !~ /^\s*$/ && $cfg{download}) {
        # download the file from the jmol website if it has been modified
        my $ua  = LWP::UserAgent->new();
           $ua->agent("$cfg{self}/$cfg{version}");

        print "Checking URL for modification: $cfg{url} ... " if($cfg{verbose});

        my $req = HTTP::Request->new( HEAD => $cfg{url},
                                      HTTP::Headers->new('If-Modified-Since' => $last_modified)
                                    );
        my $res = $ua->request($req);
        if($res->code eq '304') {
                # the file was not modified, skip download & processing
                print "Not Modified.\n" if($cfg{verbose});

                if($cfg{force}) {
                        # we will only download files if they were modified
                        # or if we're forced to

                        print "Parameter --force given, downloading anyway.\n" if($cfg{verbose});
                }
                else {
                        print "Local copy is up-to-date, skipping download.\n" if($cfg{verbose});
                        $cfg{download} = 0;
                }
        }
        else {
                print "Modified!\n" if($cfg{verbose});
        }
        print "\n" if($cfg{verbose});
}

# actually download the file
if($cfg{download} == 1) {
        my $ua  = LWP::UserAgent->new();
           $ua->agent("$cfg{self}/$cfg{version}");

        print "Downloading URL: $cfg{url} ... " if($cfg{verbose});

        my $req = HTTP::Request->new( GET => $cfg{url} );
        my $res = $ua->request($req);
        if($res->is_success) {
                print "Success\n" if($cfg{verbose});

		my ($success, $reason) = save_file($cfg{'jmol-projects-file'}, $res->content, mode => '0664');
		if(!$success) {
			print $reason;
			exit 1;
		}

		$jmol_projects = read_xml(%cfg, Name => 'jmol-projects') || die "Error: Cannot read jmol-projects XML file!\n";

                $last_modified = $res->header('Last-Modified');
                print "Last-Modified: $last_modified\n" if($cfg{verbose});
                $jmol_projects->{fah_projects}->{last_modified} = $last_modified;
                $jmol_projects->{fah_projects}->{last_update}   = get_date();

                #print Dumper($jmol_projects);
		
		print "Saving jmol-projects XML file ...\n" if($cfg{verbose});
                ($success, $reason) = write_xml( %cfg,
                                                 Name => 'jmol-projects',
                                                 XML  => $jmol_projects
                                               );
                if(!$success) {
                	print $reason;
                        exit 1;
		}
                print "\n" if($cfg{verbose});
        }
        else {
                print "Failed!\n" if($cfg{verbose});
                print "Error: Failed to retrieve URL! ($cfg{url})\n";
                print "Server returned: ".$res->code." ".$res->message."\n";

                exit 1;
        }
}


# add the data of each project Jmol knows about to the project-summary.xml if FCI doesn't already know the project too
my $updated = 0;
foreach my $project (keys %{$jmol_projects->{fah_projects}->{fah_proj}}) {
	if($project_summary->{project_summary}->{project}->{$project}) {
		# we already know the project
	}
	else {
		# Jmol knows of a project FCI does not know about yet
	
		# FCI XML example:
		#	
		# <project>
		#   <number>9999</number>
		#   <atoms>10066</atoms>
		#   <code>GRO-SMP</code>
		#   <contact>densign</contact>
		#   <credit>587.00</credit>
		#   <deadline>12.00</deadline>
		#   <description>http://fah-web.stanford.edu/cgi-bin/fahproject?p=9999</description>
		#   <frames>100</frames>
		#   <last_update>2006-12-01 13:17:08</last_update>
		#   <name>p9999_SMP-sv-melting-99p</name>
		#   <preferred>4.00</preferred>
		#   <server>171.64.65.63</server>
		#   <status>
		#     <active>false</active>
		#     <beta>true</beta>
		#     <comprehensive>true</comprehensive>
		#   </status>
		# </project>
		#
		# Jmol XML example:
		#
		# <fah_proj number="9999" 
		#           name="p9999_SMP-sv-melting-99p" 
		#           public="n" 
		#           server="171.64.65.63" 
		#           atoms="10066" 
		#           preferred="4" 
		#           deadline="12" 
		#           credit="587" 
		#           frames="100" 
		#           code="GS" 
		#           contact="densign"
		# />
		
		if(   ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{atoms}     &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{atoms}     !~ /^\s*$/) 
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}      &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}      !~ /^\s*$/) 
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{contact}   &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{contact}   !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit}    &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit}    !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{deadline}  &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{deadline}  !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{frames}    &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{frames}    !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{name}      &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{name}      !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{preferred} &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{preferred} !~ /^\s*$/)
		   &&
		      ($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{server}    &&
		       $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{server}    !~ /^\s*$/)
		) {
			# Core identifiers used by the various parties:
			#
			# Jmol		      |	psummary.html |	Core # | Core Name
			# --------------------+---------------+--------+----------------------------
			# <code>A</code>	AMBER		0x82	 Folding@Home PMD Core
			# <code>A0</code>	GROMACS33	0xa0	 Folding@Home Gromacs 3.3 Core
			# <code>AD</code>	ATI-DEV  	0x12	 ?
			# <code>DB</code>	DGROMACSB	0x7b	 Folding@Home Double Gromacs Core B
    			# <code>DC</code>	DGROMACSC	0x7c     Folding@Home Double Gromacs Core C
			# <code>DG</code>	DGROMACS	0x79	 Folding@Home Double Gromacs Core
			# <code>G3</code>	GRO-PS3		?	 ?
			# <code>GB</code>	GBGROMACS	0x7a	 Folding@Home GB Gromacs Core
			# <code>G</code>	GROMACS		0x78	 Folding@Home Gromacs Core
			# <code>GC</code>       GROCVS          0xa2     Folding@Home Gromacs SMP Core (Core Version 1.90)
			# <code>GG</code>	GROGPU		0x10	 Folding@Home GPU Core 
    			# <code>GG2</code>	GROGPU2		0x11	 Folding@home GPU2 Core
    			# <code>GG2MT</code>	GROGPU2-MT	0x14	 Folding@home GPU2 Core
			# <code>GS</code>	GRO-SMP		0xa1	 Folding@Home Gromacs SMP Core (Core Version 1.74)
			# <code>GST</code>	GROSIMT		0x81	 Folding@Home Gromacs Simulated Tempering Core
			# <code>GT</code>	GROST		0x80	 Folding@Home Gromacs SREM Core
			# <code>ND</code>	NVIDIA-DEV	0x13	 ?
			# <code>Q</code>	QMD		0x96	 Folding@Home QMD Core
			# <code>SP</code>       SHARPEN		?	 ?
			# <code>T</code>	TINKER		0x65	 Folding@Home Client Core 
                        #                                                TINKER: Software Tools for Molecular Design

			my %code = ( 
				     'A'     => 'AMBER',
				     'A0'    => 'GROMACS33',
				     'AD'    => 'ATI-DEV',
			             'DB'    => 'DGROMACSB',
			             'DC'    => 'DGROMACSC',
			             'DG'    => 'DGROMACS',
				     'G'     => 'GROMACS',
				     'G3'    => 'GRO-PS3',
			             'GB'    => 'GBGROMACS',
			             'GC'    => 'GROCVS',
				     'GG'    => 'GROGPU',
				     'GG2'   => 'GROGPU2',
				     'GG2MT' => 'GROGPU2-MT',
				     'GS'    => 'GRO-SMP',
				     'GST'   => 'GROSIMT',
				     'GT'    => 'GROST',
				     'ND'    => 'NVIDIA-DEV',
			             'Q'     => 'QMD',
			             'SP'    => 'SHARPEN',
			             'T'     => 'TINKER',
			           );

			if(!exists $code{$jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}}) {
				print "Skipping import of project $project from the Jmol project XML file.\n";
				print "No mapping exists for code: ".$jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}."\n";
				
				next;
			}
			
			$project_summary->{project_summary}->{project}->{$project}->{number} = $project;
			
			# Copy values from Jmol that can be copied 1-on-1

			my %keys = ( atoms     => 'atoms',
	                             contact   => 'contact',
				     frames    => 'frames',
				     name      => 'name',
				     server    => 'server',
                	           );
			
			foreach my $key (keys %keys) {
				if($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{$key}) {
					$project_summary->{project_summary}->{project}->{$project}->{$key} 
						= $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{$key};
				}
			}

			# Copy values from Jmol that require format modifications and other special treatments

			# deadline
			if($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{deadline} =~ /^\d+(\.\d+)?$/) {
				$project_summary->{project_summary}->{project}->{$project}->{deadline}
					= sprintf("%.2f", 
						$jmol_projects->{fah_projects}->{fah_proj}->{$project}->{deadline});
			}
			
			# preferred
			if($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{preferred} =~ /^\d+(\.\d+)?$/) {
				$project_summary->{project_summary}->{project}->{$project}->{preferred}
					= sprintf("%.2f", 
						$jmol_projects->{fah_projects}->{fah_proj}->{$project}->{preferred});
			}
			
			# credit
			if($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit} =~ /^\d+$/) {
				$project_summary->{project_summary}->{project}->{$project}->{credit}
					= $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit} .'.00';
			}
			elsif($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit} =~ /^\d+\.\d$/) {
				$project_summary->{project_summary}->{project}->{$project}->{credit}
					= $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit} .'0';
			}
			else {
				$project_summary->{project_summary}->{project}->{$project}->{credit}
					= $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{credit};
			}
			
			# code
			if($jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}) {
				$project_summary->{project_summary}->{project}->{$project}->{code} 
					= $code{$jmol_projects->{fah_projects}->{fah_proj}->{$project}->{code}};
			}
			
			# description
			$project_summary->{project_summary}->{project}->{$project}->{description}
				= "http://fah-web.stanford.edu/cgi-bin/fahproject?p=$project";
			
			# last_update
			$project_summary->{project_summary}->{project}->{$project}->{last_update}
				= get_date();
	
			$updated = 1;

			print "Imported project from Jmol: $project\n";
			foreach(sort keys %{$project_summary->{project_summary}->{project}->{$project}}) {
				print " $_: ";
				print $project_summary->{project_summary}->{project}->{$project}->{$_};
				print "\n";
			}
		}
	}
}

if($updated) {
	print "Saving project-summary.xml file ...\n" if($cfg{verbose});
	my ($success, $reason) = write_xml( %cfg, 
	                                    Name => 'xml-project-summary', 
					    XML  => $project_summary
					  );
}


exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


#############
#
#     Routine: fetch_url()
#  Parameters: user-agent object
#              url
#     Returns: page content on success or
#              undef in case of no success
# Description: retrieve a file via an HTTP GET request
# 

sub fetch_url {
	my $ua  = $_[0];
	my $url = $_[1];
	
	my $req = HTTP::Request->new(GET => $url);
	
	print "Retrieving URL: $url ... " if($cfg{verbose});
	my $res = $ua->request($req);
	if($res->is_success) {
		print "Success\n" if($cfg{verbose});
		
		return $res->content;
	}
	else {
		print "Failed!\n" if($cfg{verbose});
		print "Error: Failed to retrieve URL! ($url)\n";
		print "HTTP Status: ".$res->code." ".$res->message."\n";
			
		return undef;
	}
}
																	

##############
#
#     Routine: valid_url()
#  Parameters: url
#     Returns: 1 on success or
#              2 in case of authentication required or
#              0 otherwise
# Description: verify the access to a file via a HTTP HEAD request and
#              detect requirement of HTTP Authentication
# 

sub valid_url {
	my $url = $_[0];
	
	my $ua  = LWP::UserAgent->new();
	   $ua->agent($cfg{useragent});
	
	my $req = HTTP::Request->new(HEAD => $url);
	my $res = $ua->request($req);
	
	# store realm and http status info in global hash
	$http_info{$url}{realm}  = $res->header('WWW-Authenticate');
	$http_info{$url}{status} = $res->code." ".$res->message;

	if($res->is_success()) {
		return 1;
	}
	elsif($res->code == 401) {
		return 2;
	}
	else {
		return 0;
	}
}


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats|eoc-user-summary)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'jmol-projects') {
			return XMLin( $args{File},
                                      KeyAttr    => { fah_proj => 'number' },
                                      ForceArray => [ 'fah_proj' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'eoc-user-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
                                KeyAttr    => { fah_proj => 'number' },
				NoAttr     => 1,
			        KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'eoc-user-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
		
		# allow group users write permissions
		chmod(0664, $args{File});
	}

	return ($success, $reason);
}


sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-n, --no-download       Don't download any files, use the ones on disk\n";
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
